from typing import TypeAlias

from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.verification.core import *
from rainbowadn.inlining import *
from rainbowadn.nullability import *
from ._bankflow import *
from ._flowbank import *
from ._flowblock import *
from ._flowcheque import *
from ._flowstandard import *

__all__ = ('BankBlock',)

Link: TypeAlias = IPair[FlowCheque, FlowBank]
Block: TypeAlias = FlowBlock[Link]
Index: TypeAlias = FlowStandard[FlowBlock[Link]]


class BankBlock:
    def __init__(self, reference: NullableReference[Block]):
        assert isinstance(reference, NullableReference)
        self.reference = reference

    @classmethod
    def flow(cls) -> 'BankFlow':
        return BankFlow(FlowBank.empty())

    @classmethod
    def link_factory(cls) -> RainbowFactory[Link]:
        return IPair.f(FlowCheque.factory(), FlowBank.factory())

    @classmethod
    def empty(cls) -> 'BankBlock':
        return cls(
            NullableReference(
                Null(),
                FlowBlockFactory(cls.link_factory()).loose()
            )
        )

    @classmethod
    def verification(cls) -> Verification[Index]:
        return FlowBlockVerification(cls.flow().link_verification()).loose()

    async def verify(self) -> bool:
        assert_true(await self.verification().verify(await FlowBlock.outer_of(self.link_factory(), self.reference)))
        return True

    async def add(self, cheque: FlowCheque) -> 'BankBlock':
        assert isinstance(cheque, FlowCheque)
        return await AddCheque(cheque).map(self)


class AddCheque(
    Mapper[BankBlock, BankBlock]
):
    def __init__(self, cheque: FlowCheque):
        assert isinstance(cheque, FlowCheque)
        self.cheque = cheque

    @classmethod
    async def _bank_for_link(cls, link: Nullable[Link]) -> Nullable[FlowBank]:
        assert isinstance(link, Nullable)
        return Null() if link.null() else NotNull(await (link.resolve()).e1.resolve())

    async def _next_bank_for_bank(self, bank: Nullable[FlowBank]) -> FlowBank:
        return await BankBlock.flow().add(bank, self.cheque)

    async def _link_for_bank(self, bank: FlowBank) -> HashPoint[Link]:
        assert isinstance(bank, FlowBank)
        return HashPoint.of(IPair.of(self.cheque, bank))

    async def _next_link_for_link(
            self, link: Nullable[Link]
    ) -> HashPoint[Link]:
        assert isinstance(link, Nullable)
        return await self._link_for_bank(
            await self._next_bank_for_bank(
                await self._bank_for_link(link)
            )
        )

    @classmethod
    async def _add_link_to_reference(
            cls,
            link: HashPoint[Link],
            reference: NullableReference[Block],
    ) -> BankBlock:
        assert isinstance(link, HashPoint)
        assert isinstance(reference, NullableReference)
        return BankBlock(NullableReference.off(await FlowBlock.add_to(reference, link)))

    async def _add_to_link(
            self,
            previous: Nullable[Link],
            reference: NullableReference[Block],
    ) -> BankBlock:
        assert isinstance(previous, Nullable)
        assert isinstance(reference, NullableReference)
        return await self._add_link_to_reference(await self._next_link_for_link(previous), reference)

    async def _add_to_reference(
            self,
            reference: NullableReference[Block],
    ) -> BankBlock:
        assert isinstance(reference, NullableReference)
        return await self._add_to_link(await FlowBlock.link_of(reference), reference)

    async def map(self, element: BankBlock) -> BankBlock:
        return await self._add_to_reference(element.reference)
